"""Yale's mail server is pretty dumb.

Its reports include the end user's name, but not the full domain.  I think we
can usually guess it right anyway.  This is completely based on examination of
the corpse, and is subject to failure whenever Yale even slightly changes
their MTA. :(
"""

import re

from email.message import Message
from email.utils import getaddresses
from enum import Enum
from io import BytesIO
from typing import cast

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


scre = re.compile(b'Message not delivered to the following', re.IGNORECASE)
ecre = re.compile(b'Error Detail', re.IGNORECASE)
acre = re.compile(b'\\s+(?P<addr>\\S+)\\s+')


class ParseState(Enum):
    start = 0
    intro_found = 1


@public
class Yale:
    """Parse Yale's bounces (or what used to be)."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        if msg.is_multipart():
            return NoFailures
        try:
            whofrom = getaddresses([msg.get('from', '')])[0][1]
            if not whofrom:
                return NoFailures
            username, domain = whofrom.split('@', 1)
        except (IndexError, ValueError):
            return NoFailures
        if username.lower() != 'mailer-daemon':
            return NoFailures
        parts = domain.split('.')
        parts.reverse()
        for part1, part2 in zip(parts, ('edu', 'yale'), strict=False):
            if part1 != part2:
                return NoFailures
        # Okay, we've established that the bounce came from the mailer-daemon
        # at yale.edu.  Let's look for a name, and then guess the relevant
        # domains.
        names: set[bytes] = set()
        # msg was validated non-multipart above, so the decoded payload is
        # bytes at runtime; pyrefly sees the wider Message | bytes | Any union.
        body = BytesIO(cast(bytes, msg.get_payload(decode=True)))
        state = ParseState.start
        for line in body:
            if state is ParseState.start and scre.search(line):
                state = ParseState.intro_found
            elif state is ParseState.intro_found and ecre.search(line):
                break
            elif state is ParseState.intro_found:
                if mo := acre.search(line):
                    names.add(mo.group('addr'))
        # Now we have a bunch of names, these are either @yale.edu or
        # @cs.yale.edu.  Add them both.
        addresses: set[bytes] = set()
        for name in names:
            addresses.add(name + b'@yale.edu')
            addresses.add(name + b'@cs.yale.edu')
        return NoTemporaryFailures, frozenset(addresses)



"""LLNL's custom Sendmail bounce message."""

import re

from email.iterators import body_line_iterator
from email.message import Message

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


acre = re.compile(r',\s*(?P<addr>\S+@[^,]+),', re.IGNORECASE)


@public
class LLNL:
    """LLNL's custom Sendmail bounce message."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""

        for line in body_line_iterator(msg):
            if mo := acre.search(line):
                address = mo.group('addr').encode('us-ascii')
                return NoTemporaryFailures, frozenset([address])
        return NoFailures



"""Yahoo! has its own weird format for bounces."""

import re

from email.iterators import body_line_iterator
from email.utils import parseaddr
from email.message import Message
from enum import Enum

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


tcre = (
    re.compile(r'message\s+from\s+yahoo\.\S+', re.IGNORECASE),
    re.compile(
        r'Sorry, we were unable to deliver your message to '
        r'the following address(\(es\))?\.',
        re.IGNORECASE,
    ),
)
acre = re.compile(r'<(?P<addr>[^>]*)>:')
ecre = (
    re.compile(r'--- Original message follows'),
    re.compile(r'--- Below this line is a copy of the message'),
    re.compile(r'---------- Forwarded message ----------'),
)


class _ParseState(Enum):
    start = 0
    tag_seen = 1
    all_done = 2


@public
class Yahoo:
    """Yahoo! bounce detection."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        # Yahoo! bounces seem to have a known subject value and something
        # called an x-uidl: header, the value of which seems unimportant.
        sender = parseaddr(msg.get('from', '').lower())[1] or ''
        if not sender.startswith('mailer-daemon@yahoo'):
            return NoFailures
        addresses: set[bytes] = set()
        state = _ParseState.start
        for line in body_line_iterator(msg):
            line = line.strip()
            if state is _ParseState.start:
                for cre in tcre:
                    if cre.match(line):
                        state = _ParseState.tag_seen
                        break
            elif state is _ParseState.tag_seen:
                if mo := acre.match(line):
                    addresses.add(mo.group('addr').encode('us-ascii'))
                    continue
                for cre in ecre:
                    mo = cre.match(line)
                    if mo:
                        # We're at the end of the error response.
                        state = _ParseState.all_done
                        break
                if state is _ParseState.all_done:
                    break
        return NoTemporaryFailures, frozenset(addresses)



"""Parse bounce messages generated by Exim.

Exim adds an X-Failed-Recipients: header to bounce messages containing
an `addresslist' of failed addresses.
"""

from email.utils import getaddresses
from email.message import Message

from public import public

from flufl.bounce.interfaces import NoTemporaryFailures, Recipients


@public
class Exim:
    """Parse bounce messages generated by Exim."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        all_failed = msg.get_all('x-failed-recipients', [])
        # all_failed will contain string/unicode values, but the flufl.bounce
        # API requires these to be bytes.  We don't know the encoding, but
        # assume it must be ascii, per the relevant RFCs.
        return (
            NoTemporaryFailures,
            frozenset(address.encode('us-ascii') for name, address in getaddresses(all_failed)),
        )



"""Recognizes (some) Microsoft Exchange formats."""

import re

from email.iterators import body_line_iterator
from email.message import Message

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


scre = re.compile('did not reach the following recipient')
ecre = re.compile('MSEXCH:')
a1cre = re.compile('SMTP=(?P<addr>[^;]+); on ')
a2cre = re.compile('(?P<addr>[^ ]+) on ')


@public
class Exchange:
    """Recognizes (some) Microsoft Exchange formats."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        addresses: set[bytes] = set()
        it = body_line_iterator(msg)
        # Find the start line.
        for line in it:
            if scre.search(line):
                break
        else:
            return NoFailures
        # Search each line until we hit the end line.
        for line in it:
            if ecre.search(line):
                break
            if mo := a1cre.search(line) or a2cre.search(line):
                # For Python 3 compatibility, the API requires bytes
                addresses.add(mo.group('addr').encode('us-ascii'))
        return NoTemporaryFailures, frozenset(addresses)



from public import public

from flufl.bounce._scan import all_failures, scan_message


__version__ = '5.0'


public(
    all_failures=all_failures,
    scan_message=scan_message,
    __version__=__version__,
)



import logging

from public import public

from flufl.bounce.interfaces import BounceDetector, Recipients
from flufl.bounce._detectors import (
    aol,
    caiwireless,
    dsn,
    exchange,
    exim,
    groupwise,
    llnl,
    microsoft,
    netscape,
    postfix,
    qmail,
    simplematch,
    simplewarning,
    sina,
    smtp32,
    yahoo,
    yale,
)


log = logging.getLogger('flufl.bounce')


# Detectors are run in order of most specific/authoritative to least.
_DETECTORS: list[type[BounceDetector]] = [
    # RFC compliant DSNs.
    dsn.DSN,
    # Detectors for specific providers without heuristics.
    exim.Exim,
    sina.Sina,
    # Provider specific heuristic detectors.
    yahoo.Yahoo,
    yale.Yale,
    smtp32.SMTP32,
    postfix.Postfix,
    qmail.Qmail,
    groupwise.GroupWise,
    microsoft.Microsoft,
    caiwireless.Caiwireless,
    exchange.Exchange,
    netscape.Netscape,
    aol.AOL,
    # Generic and other heuristic detectors.
    simplematch.SimpleMatch,
    simplewarning.SimpleWarning,
    llnl.LLNL,
]


@public
def scan_message(msg) -> Recipients:
    """Detect the set of all permanently bouncing original recipients.

    :param msg: The bounce message.
    :type msg: `email.message.Message`
    :return: The set of detected original recipients.
    """
    permanent_failures: set[bytes] = set()
    for detector_class in _DETECTORS:
        log.info(f'Running detector: {detector_class}')
        try:
            temporary, permanent = detector_class().process(msg)
        except Exception:
            log.exception(f'Exception in detector: {detector_class}')
            raise
        permanent_failures.update(permanent)
        if temporary or permanent:
            break
    return frozenset(permanent_failures)


@public
def all_failures(msg) -> tuple[Recipients, Recipients]:
    """Detect the set of all bouncing original recipients.

    :param msg: The bounce message.
    :type msg: `email.message.Message`
    :return: 2-tuple of the temporary failure set and permanent failure set.
    """
    temporary_failures: set[bytes] = set()
    permanent_failures: set[bytes] = set()
    for detector_class in _DETECTORS:
        log.info(f'Running detector: {detector_class}')
        temporary, permanent = detector_class().process(msg)
        temporary_failures.update(temporary)
        permanent_failures.update(permanent)
        if temporary or permanent:
            break
    return frozenset(temporary_failures), frozenset(permanent_failures)



"""Parse mystery style generated by MTA at caiwireless.net."""

import re

from email.iterators import body_line_iterator
from email.message import Message
from enum import Enum

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


tcre = re.compile(r'the following recipients did not receive this message:', re.IGNORECASE)
acre = re.compile(r'<(?P<addr>[^>]*)>')


class ParseState(Enum):
    start = 0
    tag_seen = 1


@public
class Caiwireless:
    """Parse mystery style generated by MTA at caiwireless.net."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        if msg.get_content_type() == 'multipart/mixed':
            state = ParseState.start
            # This format thinks it's a MIME, but it really isn't.
            for line in body_line_iterator(msg):
                line = line.strip()
                if state is ParseState.start and tcre.match(line):
                    state = ParseState.tag_seen
                elif state is ParseState.tag_seen and line:
                    if mo := acre.match(line):
                        return NoTemporaryFailures, frozenset([mo.group('addr').encode('us-ascii')])
                    break
        return NoFailures



"""Parse bounce messages generated by Postfix.

This also matches something called 'Keftamail' which looks just like Postfix
bounces with the word Postfix scratched out and the word 'Keftamail' written
in in crayon.

It also matches something claiming to be 'The BNS Postfix program', and
'SMTP_Gateway'.  Everybody's gotta be different, huh?
"""

import re

from email.message import Message
from enum import Enum
from io import BytesIO
from typing import cast

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


# Are these heuristics correct or guaranteed?
pcre = re.compile(b'[ \\t]*the\\s*(bns)?\\s*(postfix|keftamail|smtp_gateway)', re.IGNORECASE)
rcre = re.compile(b'failure reason:$', re.IGNORECASE)
acre = re.compile(b'<(?P<addr>[^>]*)>:')

REPORT_TYPES = ('multipart/mixed', 'multipart/report')


class ParseState(Enum):
    start = 0
    salutation_found = 1


def findaddr(msg: Message) -> set[bytes]:
    addresses: set[bytes] = set()
    # findaddr is only called with a text/plain leaf subpart, so the decoded
    # payload is bytes at runtime; pyrefly sees the wider union.
    body = BytesIO(cast(bytes, msg.get_payload(decode=True)))
    state = ParseState.start
    for line in body:
        # Preserve leading whitespace.
        line = line.rstrip()
        # Yes, use match() to match at beginning of string.
        if state is ParseState.start and (pcre.match(line) or rcre.match(line)):
            # Then...
            state = ParseState.salutation_found
        elif state is ParseState.salutation_found and line:
            if mo := acre.search(line):
                addresses.add(mo.group('addr'))
            # Probably a continuation line.
    return addresses


@public
class Postfix:
    """Parse bounce messages generated by Postfix."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        if msg.get_content_type() not in REPORT_TYPES:
            return NoFailures
        # We're looking for the plain/text subpart with a Content-Description:
        # of 'notification'.
        leaves = [p for p in msg.walk() if not p.is_multipart()]
        for subpart in leaves:
            content_type = subpart.get_content_type()
            content_desc = subpart.get('content-description', '').lower()
            if content_type == 'text/plain' and content_desc == 'notification':
                return NoTemporaryFailures, frozenset(findaddr(subpart))
        return NoFailures



"""Recognizes a class of messages from AOL that report only Screen Name."""

import re

from email.message import Message
from email.utils import parseaddr
from typing import cast

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


scre = re.compile(b'mail to the following recipients could not be delivered')


@public
class AOL:
    """Recognizes a class of messages from AOL that report only Screen Name."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        if msg.get_content_type() != 'text/plain':
            return NoFailures
        if not parseaddr(msg.get('from', ''))[1].lower().endswith('@aol.com'):
            return NoFailures
        addresses: set[bytes] = set()
        found = False
        # The content-type was validated as text/plain above, so the decoded
        # payload is bytes at runtime; pyrefly can't follow the precondition
        # through Message.get_payload()'s wider union return type.
        for line in cast(bytes, msg.get_payload(decode=True)).splitlines():
            if scre.search(line):
                found = True
                continue
            if found:
                local = line.strip()
                if local:
                    if re.search(b'\\s', local):
                        break
                    if b'@' in local:
                        addresses.add(local)
                    else:
                        addresses.add(local + b'@aol.com')
        return NoTemporaryFailures, frozenset(addresses)



"""sina.com bounces"""

import re

from contextlib import suppress
from email.iterators import body_line_iterator
from email.message import Message

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


acre = re.compile(r'<(?P<addr>[^>]*)>')


@public
class Sina:
    """sina.com bounces"""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        if msg.get('from', '').lower() != 'mailer-daemon@sina.com':
            return NoFailures
        if not msg.is_multipart():
            return NoFailures
        # The interesting bits are in the first text/plain multipart.
        part = None
        with suppress(IndexError):
            part = msg.get_payload(0)
        if not part:
            return NoFailures
        addresses: set[bytes] = set()
        # `part` is a Message subpart at runtime (msg was validated multipart
        # above), but Message.get_payload(int)'s typeshed signature returns a
        # wider union than reality.
        #
        # pyrefly: ignore[bad-argument-type]
        for line in body_line_iterator(part):
            if mo := acre.match(line):
                addresses.add(mo.group('addr').encode('us-ascii'))
        return NoTemporaryFailures, frozenset(addresses)



"""This appears to be the format for Novell GroupWise and NTMail

X-Mailer: Novell GroupWise Internet Agent 5.5.3.1
X-Mailer: NTMail v4.30.0012
X-Mailer: Internet Mail Service (5.5.2653.19)
"""

import re

from email.message import Message
from io import BytesIO
from typing import cast

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


acre = re.compile(b'<(?P<addr>[^>]*)>')


def find_textplain(msg: Message) -> Message | None:
    if msg.get_content_type() == 'text/plain':
        return msg
    if msg.is_multipart():
        for part in msg.get_payload():
            if not isinstance(part, Message):
                continue
            if ret := find_textplain(part):
                return ret
    return None


@public
class GroupWise:
    """Parse Novell GroupWise and NTMail bounces."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""
        if msg.get_content_type() != 'multipart/mixed' or not msg['x-mailer']:
            return NoFailures
        if msg['x-mailer'][:3].lower() not in ('nov', 'ntm', 'int'):
            return NoFailures
        addresses: set[bytes] = set()
        # Find the first text/plain part in the message.
        text_plain = find_textplain(msg)
        if text_plain is None:
            return NoFailures
        # text_plain is a text/plain leaf Message at runtime, so the decoded
        # payload is bytes; pyrefly sees the wider Message | bytes | Any union.
        body = BytesIO(cast(bytes, text_plain.get_payload(decode=True)))
        for line in body:
            if mo := acre.search(line):
                addresses.add(mo.group('addr'))
            elif b'@' in line:
                i = line.find(b' ')
                if i == 0:
                    continue
                if i < 0:
                    # TODO: `line` here still has its trailing newline (BytesIO
                    # preserves line terminators), so this fallback would add
                    # addresses with `\n`/`\r\n` attached.  No test exercises
                    # this path; revisit when a real-world sample surfaces.
                    addresses.add(line)
                else:
                    addresses.add(line[:i])
        return NoTemporaryFailures, frozenset(addresses)



"""Recognizes simple heuristically delimited warnings."""

from email.message import Message

from public import public

from flufl.bounce._detectors.simplematch import SimpleMatch, _c
from flufl.bounce.interfaces import NoPermanentFailures, Recipients


# This is a list of tuples of the form
#
#     (start cre, end cre, address cre)
#
# where 'cre' means compiled regular expression, start is the line just before
# the bouncing address block, end is the line just after the bouncing address
# block, and address cre is the regexp that will recognize the addresses.  It
# must have a group called 'addr' which will contain exactly and only the
# address that bounced.
PATTERNS = [
    # pop3.pta.lia.net
    (
        _c('The address to which the message has not yet been delivered is'),
        _c('No action is required on your part'),
        _c(r'\s*(?P<addr>\S+@\S+)\s*'),
    ),
    # MessageSwitch
    (
        _c('Your message to:'),
        _c('This is just a warning, you do not need to take any action'),
        _c(r'\s*(?P<addr>\S+@\S+)\s*'),
    ),
    # Symantec_AntiVirus_for_SMTP_Gateways
    (
        _c('Your message with Subject:'),
        _c('Delivery attempts will continue to be made'),
        _c(r'\s*(?P<addr>\S+@\S+)\s*'),
    ),
    # googlemail.com warning
    (
        _c('Delivery to the following recipient has been delayed'),
        _c('Message will be retried'),
        _c(r'\s*(?P<addr>\S+@\S+)\s*'),
    ),
    # Exchange warning message.
    (_c('This is an advisory-only email'), _c('has been postponed'), _c('"(?P<addr>[^"]+)"')),
    # kundenserver.de
    (
        _c('not yet been delivered'),
        _c('No action is required on your part'),
        _c(r'\s*<?(?P<addr>\S+@[^>\s]+)>?\s*'),
    ),
    # Next one goes here...
]


@public
class SimpleWarning(SimpleMatch):
    """Recognizes simple heuristically delimited warnings."""

    PATTERNS = PATTERNS

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `SimpleMatch`."""
        # Since these are warnings, they're classified as temporary failures.
        # There are no permanent failures.
        _temporary, permanent_really_temporary = super().process(msg)
        return permanent_really_temporary, NoPermanentFailures



"""Microsoft's `SMTPSVC' nears I kin tell."""

import re

from email.message import Message
from enum import Enum
from io import BytesIO
from typing import cast

from public import public

from flufl.bounce._detectors.simplematch import VALID
from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


scre = re.compile(rb'transcript of session follows', re.IGNORECASE)


class ParseState(Enum):
    start = 0
    tag_seen = 1


@public
class Microsoft:
    """Microsoft's `SMTPSVC' nears I kin tell."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        if msg.get_content_type() != 'multipart/mixed':
            return NoFailures
        # Find the first subpart, which has no MIME type.  msg was validated
        # multipart/mixed above, so get_payload(0) returns a Message subpart
        # at runtime; pyrefly sees the wider Message | str | Any union.
        try:
            subpart = cast(Message, msg.get_payload(0))
        except IndexError:
            # The message *looked* like a multipart but wasn't.
            return NoFailures
        data = subpart.get_payload(decode=True)
        if isinstance(data, list):
            # The message is a multi-multipart, so not a matching bounce.
            return NoFailures
        # `subpart` is a Message and not multi-multipart (handled above), so
        # the decoded payload is bytes at runtime.
        body = BytesIO(cast(bytes, data))
        state = ParseState.start
        addresses: set[bytes] = set()
        for line in body:
            if state is ParseState.start:
                if scre.search(line):
                    state = ParseState.tag_seen
            elif state is ParseState.tag_seen and VALID.match(line.strip()):
                addresses.add(line.strip())
        return NoTemporaryFailures, frozenset(addresses)



"""Netscape Messaging Server bounce formats.

I've seen at least one NMS server version 3.6 (envy.gmp.usyd.edu.au) bounce
messages of this format.  Bounces come in DSN MIME format, but don't include
any -Recipient: headers.  Gotta just parse the text :(

NMS 4.1 (dfw-smtpin1.email.verio.net) seems even worse, but we'll try to
decipher the format here too.

"""

import re

from email.message import Message
from io import BytesIO
from typing import cast

from public import public

from flufl.bounce.interfaces import NoFailures, NoTemporaryFailures, Recipients


pcre = re.compile(b'This Message was undeliverable due to the following reason:', re.IGNORECASE)

acre = re.compile(b'(?P<reply>please reply to)?.*<(?P<addr>[^>]*)>', re.IGNORECASE)


@public
class Netscape:
    """Netscape Messaging Server bounce formats."""

    def process(self, msg: Message) -> tuple[Recipients, Recipients]:
        """See `BounceDetector`."""

        # Sigh.  Some NMS 3.6's show
        #     multipart/report; report-type=delivery-status
        # and some show
        #     multipart/mixed;
        if not msg.is_multipart():
            return NoFailures
        # We're looking for a text/plain subpart occuring before a
        # message/delivery-status subpart.
        plainmsg = None
        leaves = [p for p in msg.walk() if not p.is_multipart()]
        for _i, subpart in zip(range(len(leaves) - 1), leaves, strict=False):
            if subpart.get_content_type() == 'text/plain':
                plainmsg = subpart
                break
        if not plainmsg:
            return NoFailures
        # Total guesswork, based on captured examples...
        # plainmsg is the selected text/plain leaf, so the decoded payload is
        # bytes at runtime; pyrefly sees the wider Message | bytes | Any union.
        body = BytesIO(cast(bytes, plainmsg.get_payload(decode=True)))
        addresses: set[bytes] = set()
        for line in body:
            if mo := pcre.search(line):
                # We found a bounce section, but I have no idea what the
                # official format inside here is.  :( We'll just search for
                # <addr> strings.
                for line in body:
                    if (mo := acre.search(line)) and not mo.group('reply'):
                        addresses.add(mo.group('addr'))
        return NoTemporaryFailures, frozenset(addresses)
